from copy import copy

from dateutil.relativedelta import relativedelta
from edc_constants.constants import MALE, YES
from edc_form_validators import FormValidator
from edc_reportable import ReportablesFormValidatorMixin
from edc_utils.date import get_utcnow


class SpecimenResultFormValidator(ReportablesFormValidatorMixin, FormValidator):

    reference_range_collection_name = "my_reference_list"

    def clean(self):
        self.validate_reportable_fields()

    def validate_reportable_fields(self, *args, **kwargs):

        reportables = self.reportables_cls(
            self.reference_range_collection_name,
            cleaned_data=copy(self.cleaned_data),
            gender=MALE,
            dob=get_utcnow() - relativedelta(years=25),
            report_datetime=get_utcnow(),
        )
        reportables.validate_reportable_fields()

        reportables.validate_results_abnormal_field()

        self.applicable_if(
            YES, field="results_abnormal", field_applicable="results_reportable"
        )

        reportables.validate_results_reportable_field()



from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_constants.constants import BLACK, MALE
from edc_lab import site_labs
from edc_lab.models import Panel
from edc_lab_panel.panels import rft_panel
from edc_registration.models import RegisteredSubject
from edc_reportable import (
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    site_reportables,
)
from edc_reportable.grading_data.daids_july_2017 import grading_data
from edc_reportable.normal_data.africa import normal_data
from edc_utils import get_utcnow
from edc_utils.round_up import round_half_away_from_zero

from edc_egfr.calculators import EgfrCalculatorError
from edc_egfr.egfr import Egfr, EgfrError
from egfr_app.lab_profiles import lab_profile
from egfr_app.models import (
    Appointment,
    EgfrDropNotification,
    ResultCrf,
    SubjectRequisition,
    SubjectVisit,
)


class TestEgfr(TestCase):
    def setUp(self) -> None:
        RegisteredSubject.objects.create(
            subject_identifier="1234",
            gender=MALE,
            dob=get_utcnow() - relativedelta(years=30),
            ethnicity=BLACK,
        )

    @classmethod
    def setUpTestData(cls):
        site_reportables._registry = {}
        site_reportables.register(
            name="my_reference_list", normal_data=normal_data, grading_data=grading_data
        )
        site_labs.initialize()
        site_labs.register(lab_profile=lab_profile)

    def test_ok(self):
        egfr = Egfr(
            gender=MALE,
            age_in_years=30,
            ethnicity=BLACK,
            creatinine_value=52.0,
            creatinine_units=MICROMOLES_PER_LITER,
            report_datetime=get_utcnow(),
            reference_range_collection_name="my_reference_list",
            formula_name="ckd-epi",
        )

        try:
            self.assertGreater(egfr.egfr_value, 0.0)
        except EgfrCalculatorError as e:
            self.fail(e)

        try:
            self.assertIsNone(egfr.egfr_grade)
        except EgfrCalculatorError as e:
            self.fail(e)

        try:
            self.assertGreaterEqual(egfr.egfr_drop_value, 0.0)
        except EgfrCalculatorError as e:
            self.fail(e)

        try:
            self.assertIsNone(egfr.egfr_drop_grade)
        except EgfrCalculatorError as e:
            self.fail(e)

    def test_egfr_invalid_calculator(self):
        self.assertRaises(
            EgfrError,
            Egfr,
            gender=MALE,
            age_in_years=25,
            ethnicity=BLACK,
            creatinine_value=10.15,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
            report_datetime=get_utcnow(),
            reference_range_collection_name="my_reference_list",
        )

    def test_egfr_missing_age_and_dob_raises(self):
        self.assertRaises(
            EgfrError,
            Egfr,
            gender=MALE,
            ethnicity=BLACK,
            creatinine_value=10.15,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
            report_datetime=get_utcnow(),
            reference_range_collection_name="my_reference_list",
            formula_name="ckd-epi",
        )

    def test_egfr_percent_drop_less_than_one_raises(self):
        self.assertRaises(
            EgfrError,
            Egfr,
            gender=MALE,
            age_in_years=30,
            ethnicity=BLACK,
            creatinine_value=10.15,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
            report_datetime=get_utcnow(),
            percent_drop_threshold=0.25,
            reference_range_collection_name="my_reference_list",
            formula_name="ckd-epi",
        )

    def test_egfr_assay_datetime(self):
        self.assertRaises(
            EgfrError,
            Egfr,
            gender=MALE,
            age_in_years=30,
            ethnicity=BLACK,
            creatinine_value=10.15,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
            report_datetime=get_utcnow(),
            assay_datetime=get_utcnow(),
            percent_drop_threshold=0.25,
            reference_range_collection_name="my_reference_list",
            formula_name="ckd-epi",
        )

    def test_egfr_grade(self):
        egfr = Egfr(
            gender=MALE,
            age_in_years=30,
            ethnicity=BLACK,
            creatinine_value=275,
            creatinine_units=MICROMOLES_PER_LITER,
            report_datetime=get_utcnow(),
            reference_range_collection_name="my_reference_list",
            formula_name="ckd-epi",
        )

        self.assertEqual(egfr.egfr_grade, 4)

    def test_egfr_dob(self):
        egfr = Egfr(
            gender=MALE,
            dob=get_utcnow() - relativedelta(years=30),
            ethnicity=BLACK,
            creatinine_value=275,
            creatinine_units=MICROMOLES_PER_LITER,
            report_datetime=get_utcnow(),
            reference_range_collection_name="my_reference_list",
            formula_name="ckd-epi",
        )
        self.assertEqual(egfr.egfr_grade, 4)

    def test_egfr_drop(self):
        opts = dict(
            gender=MALE,
            age_in_years=25,
            ethnicity=BLACK,
            creatinine_value=10.15,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
            report_datetime=get_utcnow(),
            reference_range_collection_name="my_reference_list",
            formula_name="ckd-epi",
        )
        egfr = Egfr(**opts)
        self.assertEqual(egfr.egfr_drop_value, 0.0)
        egfr = Egfr(baseline_egfr_value=23.0, **opts)
        self.assertEqual(round_half_away_from_zero(egfr.egfr_value, 2), 7.33)
        self.assertEqual(egfr.egfr_grade, 4)
        self.assertEqual(egfr.egfr_grade, 4)
        self.assertEqual(round_half_away_from_zero(egfr.egfr_drop_value, 2), 68.15)
        self.assertEqual(egfr.egfr_drop_grade, 4)
        self.assertEqual(egfr.egfr_drop_grade, 4)

    @override_settings(EDC_EGFR_DROP_NOTIFICATION_MODEL="egfr_app.EgfrDropNotification")
    def test_egfr_drop_with_notify(self):
        appointment = Appointment.objects.create(
            subject_identifier="1234",
            appt_datetime=get_utcnow(),
            timepoint=0,
        )
        subject_visit = SubjectVisit.objects.create(
            subject_identifier="1234",
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
        )

        panel = Panel.objects.get(name=rft_panel.name)

        requisition = SubjectRequisition.objects.create(
            subject_identifier="1234",
            subject_visit=subject_visit,
            report_datetime=appointment.appt_datetime,
            panel=panel,
        )
        crf = ResultCrf.objects.create(
            subject_visit=subject_visit,
            requisition=requisition,
            report_datetime=appointment.appt_datetime,
            assay_datetime=appointment.appt_datetime,
            egfr_value=156.43,
            creatinine_value=53,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        opts = dict(
            gender=MALE,
            age_in_years=30,
            ethnicity=BLACK,
            report_datetime=get_utcnow(),
            reference_range_collection_name="my_reference_list",
            formula_name="ckd-epi",
        )

        egfr = Egfr(
            baseline_egfr_value=220.1, percent_drop_threshold=20, calling_crf=crf, **opts
        )
        self.assertEqual(round_half_away_from_zero(egfr.egfr_value, 2), 156.43)
        self.assertIsNone(egfr.egfr_grade)
        self.assertEqual(round_half_away_from_zero(egfr.egfr_drop_value, 2), 28.93)
        self.assertEqual(egfr.egfr_drop_grade, 2)
        self.assertTrue(
            EgfrDropNotification.objects.filter(subject_visit=subject_visit).exists()
        )

        crf.creatinine_value = 48
        crf.save()
        crf.refresh_from_db()
        egfr = Egfr(
            baseline_egfr_value=220.1, percent_drop_threshold=20, calling_crf=crf, **opts
        )
        self.assertEqual(round_half_away_from_zero(egfr.egfr_value, 2), 162.93)
        self.assertIsNone(egfr.egfr_grade)
        self.assertEqual(round_half_away_from_zero(egfr.egfr_drop_value, 2), 25.97)
        self.assertEqual(egfr.egfr_drop_grade, 2)
        self.assertTrue(
            EgfrDropNotification.objects.filter(subject_visit=subject_visit).exists()
        )

        EgfrDropNotification.objects.all().delete()

        crf.creatinine_value = 53
        crf.save()
        crf.refresh_from_db()
        egfr = Egfr(
            baseline_egfr_value=190.1, percent_drop_threshold=20, calling_crf=crf, **opts
        )
        self.assertEqual(round_half_away_from_zero(egfr.egfr_value, 2), 156.43)
        self.assertIsNone(egfr.egfr_grade)
        self.assertEqual(round_half_away_from_zero(egfr.egfr_drop_value, 2), 17.71)
        self.assertEqual(egfr.egfr_drop_grade, 2)
        self.assertEqual(egfr.egfr_drop_grade, 2)
        self.assertFalse(
            EgfrDropNotification.objects.filter(subject_visit=subject_visit).exists()
        )

        egfr = Egfr(
            baseline_egfr_value=100.1, percent_drop_threshold=20, calling_crf=crf, **opts
        )
        self.assertEqual(round_half_away_from_zero(egfr.egfr_value, 2), 156.43)
        self.assertIsNone(egfr.egfr_grade)
        self.assertEqual(round_half_away_from_zero(egfr.egfr_drop_value, 2), 0.0)
        self.assertIsNone(egfr.egfr_drop_grade)
        self.assertFalse(
            EgfrDropNotification.objects.filter(subject_visit=subject_visit).exists()
        )



from __future__ import annotations

from decimal import Decimal

from edc_constants.constants import FEMALE
from edc_reportable import MICROMOLES_PER_LITER

from .base_egrfr import BaseEgfr, EgfrCalculatorError


class EgfrCockcroftGault(BaseEgfr):

    """Reference https://www.mdcalc.com/creatinine-clearance-cockcroft-gault-equation

    Cockcroft-Gault

    eGFR (mL/min) = { (140 – age (years)) x weight (kg) x constant*} /
    serum creatinine (μmol/L)
    *constant = 1.23 for males and 1.05 for females

    Cockcroft-Gault CrCl, mL/min =
        (140 – age) × (weight, kg) × (0.85 if female) / (72 × SCr(mg/dL))

    or:

    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2763564/

    GFR = 141 × min(Scr/κ, 1)α × max(Scr/κ, 1)-1.209 × 0.993Age
    """

    def __init__(self, weight: int | float | Decimal = None, **kwargs):
        self.weight = float(weight) if weight else None
        super().__init__(**kwargs)

    @property
    def value(self) -> float:
        """Returns the eGFR value or raises.

        eGFR (mL/min) = { (140 – age (years)) x weight (kg) x constant*} /
        serum creatinine (μmol/L)

        *constant = 1.23 for males and 1.05 for females
        """
        if (
            self.gender
            and self.age_in_years
            and self.weight
            and self.scr.get(MICROMOLES_PER_LITER)
        ):
            gender_factor = 1.05 if self.gender == FEMALE else 1.23
            adjusted_age = 140.00 - self.age_in_years
            value = (adjusted_age * self.weight * gender_factor) / float(
                self.scr.get(MICROMOLES_PER_LITER)
            )
            return value
        opts = dict(
            gender=self.gender,
            age_in_years=self.age_in_years,
            weight=self.weight,
            scr=self.scr.get(MICROMOLES_PER_LITER),
        )
        raise EgfrCalculatorError(
            f"Unable to calculate egfr_value. Insufficient information. Got {opts}."
        )



from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_constants.constants import BLACK, CLOSED, COMPLETE, INCOMPLETE, MALE, OPEN
from edc_lab import site_labs
from edc_lab.models import Panel
from edc_lab_panel.panels import rft_panel
from edc_registration.models import RegisteredSubject
from edc_reportable import MICROMOLES_PER_LITER, site_reportables
from edc_reportable.grading_data.daids_july_2017 import grading_data
from edc_reportable.normal_data.africa import normal_data
from edc_utils import get_utcnow

from edc_egfr.egfr import Egfr
from egfr_app.lab_profiles import lab_profile
from egfr_app.models import (
    Appointment,
    EgfrDropNotification,
    ResultCrf,
    SubjectRequisition,
    SubjectVisit,
)


class TestEgfr(TestCase):
    def setUp(self) -> None:
        RegisteredSubject.objects.create(
            subject_identifier="1234",
            gender=MALE,
            dob=get_utcnow() - relativedelta(years=30),
            ethnicity=BLACK,
        )
        appointment = Appointment.objects.create(
            subject_identifier="1234",
            appt_datetime=get_utcnow(),
            timepoint=0,
        )
        self.subject_visit = SubjectVisit.objects.create(
            subject_identifier="1234",
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
        )

        panel = Panel.objects.get(name=rft_panel.name)

        requisition = SubjectRequisition.objects.create(
            subject_identifier="1234",
            subject_visit=self.subject_visit,
            report_datetime=appointment.appt_datetime,
            panel=panel,
        )
        self.crf = ResultCrf.objects.create(
            subject_visit=self.subject_visit,
            requisition=requisition,
            report_datetime=appointment.appt_datetime,
            assay_datetime=appointment.appt_datetime,
            egfr_value=156.43,
            creatinine_value=53,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.opts = dict(
            gender=MALE,
            age_in_years=30,
            ethnicity=BLACK,
            report_datetime=get_utcnow(),
            reference_range_collection_name="my_reference_list",
            formula_name="ckd-epi",
        )

    @classmethod
    def setUpTestData(cls):
        site_reportables._registry = {}
        site_reportables.register(
            name="my_reference_list", normal_data=normal_data, grading_data=grading_data
        )
        site_labs.initialize()
        site_labs.register(lab_profile=lab_profile)

    @override_settings(EDC_EGFR_DROP_NOTIFICATION_MODEL="egfr_app.EgfrDropNotification")
    def test_egfr_drop_notification_model(self):
        Egfr(
            baseline_egfr_value=220.1,
            percent_drop_threshold=20,
            calling_crf=self.crf,
            **self.opts,
        )
        obj = EgfrDropNotification.objects.get(subject_visit=self.subject_visit)
        obj.report_status = OPEN
        obj.save()
        self.assertEqual(obj.crf_status, INCOMPLETE)
        obj.report_status = CLOSED
        obj.save()
        self.assertEqual(obj.crf_status, COMPLETE)



from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal
from typing import Any, Optional
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from edc_constants.constants import NEW
from edc_reportable import site_reportables
from edc_reportable.units import EGFR_UNITS, PERCENT
from edc_utils import age, get_utcnow

from .calculators import EgfrCkdEpi, EgfrCockcroftGault, egfr_percent_change
from .get_drop_notification_model import get_egfr_drop_notification_model_cls


class EgfrError(Exception):
    pass


class Egfr:

    calculators: dict = {"ckd-epi": EgfrCkdEpi, "cockcroft-gault": EgfrCockcroftGault}

    def __init__(
        self,
        baseline_egfr_value: Decimal | float | None = None,
        gender: str | None = None,
        ethnicity: str | None = None,
        age_in_years: int | None = None,
        dob: date | None = None,
        weight_in_kgs: Decimal | float | None = None,
        report_datetime: datetime | None = None,
        creatinine_value: Decimal | float | None = None,
        creatinine_units: str | None = None,
        formula_name: str | None = None,
        value_threshold: Decimal | float | None = None,
        percent_drop_threshold: Decimal | float | None = None,
        reference_range_collection_name: str | None = None,
        calling_crf: Any | None = None,
        subject_visit: Any | None = None,
        assay_datetime: datetime | None = None,
        egfr_drop_notification_model: str | None = None,
    ):
        self._egfr_value: Decimal | None = None
        self._egfr_grade = None
        self._egfr_drop_value: Decimal | None = None
        self._egfr_drop_grade = None
        self.assay_date: date | None = None
        self.subject_visit = None

        self.baseline_egfr_value = baseline_egfr_value
        if formula_name not in self.calculators:
            raise EgfrError(
                f"Invalid formula_name. Expected one of {list(self.calculators.keys())}. "
                f"Got {formula_name}."
            )
        else:
            self.calculator_cls = self.calculators.get(formula_name)
        self.age_in_years = age_in_years
        self.dob = dob
        self.weight_in_kgs = weight_in_kgs
        self.egfr_drop_notification_model = egfr_drop_notification_model
        self.egfr_drop_units = PERCENT
        self.egfr_units = EGFR_UNITS
        self.ethnicity = ethnicity
        self.gender = gender
        self.reference_range_collection_name = reference_range_collection_name
        self.report_datetime = report_datetime
        self.value_threshold = value_threshold
        self.percent_drop_threshold = percent_drop_threshold

        if self.dob:
            self.age_in_years = age(
                born=self.dob, reference_dt=self.report_datetime.astimezone(ZoneInfo("UTC"))
            ).years
        elif not self.dob and self.age_in_years:
            self.dob = (
                self.report_datetime.astimezone(ZoneInfo("UTC"))
                - relativedelta(years=self.age_in_years)
            ).date()
        else:
            raise EgfrError("Expected `age_in_years` or `dob`. Got None for both.")

        if calling_crf:
            self.creatinine_units = calling_crf.creatinine_units
            self.creatinine_value = calling_crf.creatinine_value
            self.percent_drop_threshold = calling_crf.percent_drop_threshold
            self.subject_visit = calling_crf.subject_visit
            self.report_datetime = calling_crf.report_datetime
            self.assay_date = calling_crf.assay_datetime.astimezone(ZoneInfo("UTC")).date()
        else:
            self.creatinine_units = creatinine_units
            self.creatinine_value = creatinine_value
            self.percent_drop_threshold = percent_drop_threshold
            self.subject_visit = subject_visit
            self.report_datetime = report_datetime
            if assay_datetime:
                self.assay_date = assay_datetime.astimezone(ZoneInfo("UTC")).date()

        if self.percent_drop_threshold is not None and self.percent_drop_threshold < 1.0:
            raise EgfrError(
                "Attr `percent_drop_threshold` should be a percentage. "
                f"Got {self.percent_drop_threshold}"
            )

        self.on_value_threshold_reached()

        self.on_percent_drop_threshold_reached()

    def on_value_threshold_reached(self) -> None:
        """A hook to respond if egfr value is at or beyond the value
        threshold.
        """
        pass

    def on_percent_drop_threshold_reached(self) -> None:
        """A hook to respond if egfr percent drop from baseline
        is at or beyond the percent drop threshold.
        """
        if self.egfr_drop_value and self.percent_drop_threshold is not None:
            if self.egfr_drop_value >= self.percent_drop_threshold:
                self.create_or_update_egfr_drop_notification()

    @property
    def egfr_value(self) -> float:
        if self._egfr_value is None:
            self._egfr_value = self.calculator_cls(
                gender=self.gender,
                ethnicity=self.ethnicity,
                age_in_years=self.age_in_years,
                creatinine_value=self.creatinine_value,
                creatinine_units=self.creatinine_units,
                weight=self.get_weight_in_kgs(),
            ).value
        return self._egfr_value

    @property
    def egfr_grade(self) -> Optional[int]:
        if self._egfr_grade is None:
            reference_grp = site_reportables.get(self.reference_range_collection_name).get(
                "egfr"
            )
            grade_obj = reference_grp.get_grade(
                self.egfr_value,
                gender=self.gender,
                dob=self.dob,
                report_datetime=self.report_datetime,
                units=self.egfr_units,
            )
            if grade_obj:
                self._egfr_grade = grade_obj.grade
        return self._egfr_grade

    @property
    def egfr_drop_value(self) -> float:
        if self._egfr_drop_value is None:
            if self.baseline_egfr_value:
                egfr_drop_value = egfr_percent_change(
                    float(self.egfr_value), float(self.baseline_egfr_value)
                )
            else:
                egfr_drop_value = 0.0000
            self._egfr_drop_value = 0.0000 if egfr_drop_value < 0.0000 else egfr_drop_value
        return self._egfr_drop_value

    @property
    def egfr_drop_grade(self) -> Optional[int]:
        if self._egfr_drop_grade is None:
            reference_grp = site_reportables.get(self.reference_range_collection_name).get(
                "egfr_drop"
            )
            grade_obj = reference_grp.get_grade(
                self.egfr_drop_value,
                gender=self.gender,
                dob=self.dob,
                report_datetime=self.report_datetime,
                units=self.egfr_drop_units,
            )
            if grade_obj:
                self._egfr_drop_grade = grade_obj.grade
        return self._egfr_drop_grade

    def get_weight_in_kgs(self) -> Optional[float]:
        return self.weight_in_kgs

    def create_or_update_egfr_drop_notification(self):
        """Creates or updates the `eGFR notification model`"""
        with transaction.atomic():
            try:
                obj = self.egfr_drop_notification_model_cls.objects.get(
                    subject_visit__id=self.subject_visit.id
                )
            except ObjectDoesNotExist:
                obj = self.egfr_drop_notification_model_cls.objects.create(
                    subject_visit_id=self.subject_visit.id,
                    report_datetime=self.report_datetime,
                    egfr_value=self.egfr_value,
                    creatinine_date=self.assay_date,
                    creatinine_value=self.creatinine_value,
                    creatinine_units=self.creatinine_units,
                    weight=self.get_weight_in_kgs(),
                    egfr_percent_change=self.egfr_drop_value,
                    report_status=NEW,
                    consent_version=self.subject_visit.consent_version,
                    site_id=self.subject_visit.site.id,
                )
            else:
                obj.egfr_value = self.egfr_value
                obj.creatinine_value = self.creatinine_value
                obj.weight = self.get_weight_in_kgs()
                obj.egfr_percent_change = self.egfr_drop_value
                obj.creatinine_date = self.assay_date
                obj.modified = get_utcnow()
                obj.save()
        obj.refresh_from_db()
        return obj

    @property
    def egfr_drop_notification_model_cls(self):
        return get_egfr_drop_notification_model_cls()



from django import forms
from django.test import TestCase
from edc_constants.constants import BLACK, FEMALE, MALE, NON_BLACK
from edc_form_validators import FormValidator
from edc_reportable import convert_units
from edc_reportable.units import (
    GRAMS_PER_DECILITER,
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
)
from edc_utils.round_up import round_half_away_from_zero

from edc_egfr.calculators import (
    EgfrCalculatorError,
    EgfrCkdEpi,
    EgfrCockcroftGault,
    egfr_percent_change,
)
from edc_egfr.form_validator_mixins import (
    EgfrCkdEpiFormValidatorMixin,
    EgfrCockcroftGaultFormValidatorMixin,
)


class TestCalculators(TestCase):
    def test_creatinine_units(self):
        """U.S. units: 0.84 to 1.21 milligrams per deciliter (mg/dL);
        European units: 74.3 to 107 micromoles per liter (umol/L)
        """
        self.assertEqual(
            round_half_away_from_zero(
                convert_units(
                    float(0.84),
                    units_from=MILLIGRAMS_PER_DECILITER,
                    units_to=MICROMOLES_PER_LITER,
                ),
                1,
            ),
            74.3,
        )
        self.assertEqual(
            round_half_away_from_zero(
                convert_units(
                    float(1.21),
                    units_from=MILLIGRAMS_PER_DECILITER,
                    units_to=MICROMOLES_PER_LITER,
                ),
                1,
            ),
            107.0,
        )
        self.assertEqual(
            round_half_away_from_zero(
                convert_units(
                    float(74.3),
                    units_from=MICROMOLES_PER_LITER,
                    units_to=MILLIGRAMS_PER_DECILITER,
                ),
                2,
            ),
            0.84,
        )
        self.assertEqual(
            round_half_away_from_zero(
                convert_units(
                    float(107.0),
                    units_from=MICROMOLES_PER_LITER,
                    units_to=MILLIGRAMS_PER_DECILITER,
                ),
                2,
            ),
            1.21,
        )

    def test_egfr_ckd_epi_calculator(self):

        # raises on invalid gender
        self.assertRaises(
            EgfrCalculatorError,
            EgfrCkdEpi,
            gender="blah",
            age_in_years=30,
            creatinine_value=1.0,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
        )

        # raises on low age
        self.assertRaises(
            EgfrCalculatorError,
            EgfrCkdEpi,
            gender=FEMALE,
            age_in_years=3,
            creatinine_value=1.0,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
        )

        egfr = EgfrCkdEpi(
            gender=FEMALE,
            age_in_years=30,
        )
        self.assertRaises(EgfrCalculatorError, getattr, egfr, "value")

        egfr = EgfrCkdEpi(
            gender=FEMALE,
            age_in_years=30,
            creatinine_value=52.0,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(0.7, egfr.kappa)

        egfr = EgfrCkdEpi(
            gender=MALE,
            age_in_years=30,
            creatinine_value=52.0,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(0.9, egfr.kappa)

        egfr = EgfrCkdEpi(
            gender=FEMALE,
            age_in_years=30,
            creatinine_value=53.0,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(-0.329, egfr.alpha)

        egfr = EgfrCkdEpi(
            gender=MALE,
            age_in_years=30,
            creatinine_value=53.0,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(-0.411, egfr.alpha)

        egfr1 = EgfrCkdEpi(
            gender=MALE,
            ethnicity=BLACK,
            creatinine_value=53.0,
            age_in_years=30,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(round_half_away_from_zero(egfr1.value, 2), 156.43)

        egfr2 = EgfrCkdEpi(
            gender=FEMALE,
            ethnicity=BLACK,
            creatinine_value=53.0,
            age_in_years=30,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(round_half_away_from_zero(egfr2.value, 2), 141.81)

        egfr1 = EgfrCkdEpi(
            gender=MALE,
            ethnicity=NON_BLACK,
            creatinine_value=53.0,
            age_in_years=30,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(round_half_away_from_zero(egfr1.value, 2), 134.97)

        egfr2 = EgfrCkdEpi(
            gender=FEMALE,
            ethnicity=NON_BLACK,
            creatinine_value=53.0,
            age_in_years=30,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(round_half_away_from_zero(egfr2.value, 2), 122.35)

        egfr3 = EgfrCkdEpi(
            gender=MALE,
            ethnicity=BLACK,
            creatinine_value=150.8,
            age_in_years=60,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(round_half_away_from_zero(egfr3.value, 4), 49.5026)

        egfr4 = EgfrCkdEpi(
            gender=MALE,
            ethnicity=BLACK,
            creatinine_value=152.0,
            age_in_years=60,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(round_half_away_from_zero(egfr4.value, 4), 49.0295)

        egfr4 = EgfrCkdEpi(
            gender=MALE,
            ethnicity=BLACK,
            creatinine_value=152.0,
            age_in_years=59,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(round_half_away_from_zero(egfr4.value, 4), 49.3751)

        egfr = EgfrCkdEpi(
            gender=FEMALE,
            ethnicity=BLACK,
            creatinine_value=150.8,
            age_in_years=60,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(round_half_away_from_zero(egfr.value, 4), 37.1895)

        egfr = EgfrCkdEpi(
            gender=FEMALE,
            ethnicity=BLACK,
            creatinine_value=152.0,
            age_in_years=60,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertEqual(round_half_away_from_zero(egfr.value, 4), 36.8341)

    def test_egfr_cockcroft_gault_calculator(self):

        # raises on invalid gender
        self.assertRaises(
            EgfrCalculatorError,
            EgfrCockcroftGault,
            gender="blah",
            age_in_years=30,
            creatinine_value=1.0,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        # raises on low age
        self.assertRaises(
            EgfrCalculatorError,
            EgfrCockcroftGault,
            gender=FEMALE,
            age_in_years=3,
            creatinine_value=1.0,
            creatinine_units=MICROMOLES_PER_LITER,
            weight=65.0,
        )

        # raises on missing weight
        egfr = EgfrCockcroftGault(
            gender=FEMALE,
            age_in_years=30,
            creatinine_value=1.0,
            creatinine_units=MICROMOLES_PER_LITER,
        )
        self.assertRaises(EgfrCalculatorError, getattr, egfr, "value")

        egfr = EgfrCockcroftGault(
            gender=MALE,
            age_in_years=30,
            creatinine_value=50.0,
            creatinine_units=MICROMOLES_PER_LITER,
            weight=65.0,
        )
        self.assertEqual(round_half_away_from_zero(egfr.value, 2), 175.89)

        egfr = EgfrCockcroftGault(
            gender=MALE,
            age_in_years=30,
            creatinine_value=50.8,
            creatinine_units=MICROMOLES_PER_LITER,
            weight=65.0,
        )
        self.assertEqual(round_half_away_from_zero(egfr.value, 2), 173.12)

        egfr = EgfrCockcroftGault(
            gender=MALE,
            age_in_years=30,
            creatinine_value=50.9,
            creatinine_units=MICROMOLES_PER_LITER,
            weight=65.0,
        )
        self.assertEqual(round_half_away_from_zero(egfr.value, 2), 172.78)

        egfr = EgfrCockcroftGault(
            gender=FEMALE,
            age_in_years=30,
            creatinine_value=50.9,
            creatinine_units=MICROMOLES_PER_LITER,
            weight=65.0,
        )
        self.assertEqual(round_half_away_from_zero(egfr.value, 2), 147.5)

        egfr = EgfrCockcroftGault(
            gender=FEMALE,
            creatinine_value=1.3,
            age_in_years=30,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
            weight=65.0,
        )

        self.assertEqual(round_half_away_from_zero(egfr.value, 2), 65.31)

        egfr2 = EgfrCockcroftGault(
            gender=MALE,
            creatinine_value=0.9,
            age_in_years=30,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
            weight=65.0,
        )

        self.assertEqual(round_half_away_from_zero(egfr2.value, 2), 110.51)

    def test_egfr_ckd_epi_form_validator(self):
        data = dict(
            gender=MALE,
            ethnicity=BLACK,
            age_in_years=30,
        )

        class EgfrFormValidator(EgfrCkdEpiFormValidatorMixin, FormValidator):
            pass

        # not enough data
        form_validator = EgfrFormValidator(cleaned_data=data)
        egfr = form_validator.validate_egfr()
        self.assertIsNone(egfr)

        # calculates
        data.update(creatinine_value=1.3, creatinine_units=MICROMOLES_PER_LITER)
        form_validator = EgfrFormValidator(cleaned_data=data)
        egfr = form_validator.validate_egfr()
        self.assertEqual(round_half_away_from_zero(egfr, 2), 718.14)

        # calculation error: bad units
        data.update(creatinine_units=GRAMS_PER_DECILITER)
        form_validator = EgfrFormValidator(cleaned_data=data)
        self.assertRaises(forms.ValidationError, form_validator.validate_egfr)

    def test_egfr_cockcroft_gault_form_validator(self):
        data = dict(
            gender=MALE,
            weight=72,
            age_in_years=30,
        )

        class EgfrFormValidator(EgfrCockcroftGaultFormValidatorMixin, FormValidator):
            pass

        # not enough data
        form_validator = EgfrFormValidator(cleaned_data=data)
        egfr = form_validator.validate_egfr()
        self.assertIsNone(egfr)

        # calculation error: bad units
        data.update(creatinine_value=1.3, creatinine_units=GRAMS_PER_DECILITER)
        form_validator = EgfrFormValidator(cleaned_data=data)
        self.assertRaises(forms.ValidationError, form_validator.validate_egfr)

        # calculates
        data.update(creatinine_value=1.30, creatinine_units=MILLIGRAMS_PER_DECILITER)
        form_validator = EgfrFormValidator(cleaned_data=data)
        egfr = form_validator.validate_egfr()
        self.assertEqual(round_half_away_from_zero(egfr, 2), 84.75)

        # calculates
        data.update(creatinine_value=114.94, creatinine_units=MICROMOLES_PER_LITER)
        form_validator = EgfrFormValidator(cleaned_data=data)
        egfr = form_validator.validate_egfr()
        self.assertEqual(round_half_away_from_zero(egfr, 2), 84.75)

    def test_egfr_percent_change(self):
        self.assertGreater(egfr_percent_change(51.10, 131.50), 20.0)
        self.assertLess(egfr_percent_change(51.10, 61.10), 20.0)
        self.assertEqual(egfr_percent_change(51.10, 51.10), 0.0)
        self.assertLess(egfr_percent_change(51.10, 21.10), 20.0)
        self.assertEqual(egfr_percent_change(51.10, 0), 0.0)



from edc_constants.constants import FEMALE, MALE
from edc_reportable import (
    GRADE3,
    GRADE4,
    GRAMS_PER_DECILITER,
    IU_LITER,
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    MILLIMOLES_PER_LITER,
    TEN_X_9_PER_LITER,
)
from edc_reportable import parse as p

age_opts = dict(age_lower=18, age_upper=None, age_units="years", age_lower_inclusive=True)

normal_data = {
    "haemoglobin": [
        p("13.5<=x<=17.5", units=GRAMS_PER_DECILITER, gender=[MALE], **age_opts),
        p("12.0<=x<=15.5", units=GRAMS_PER_DECILITER, gender=[FEMALE], **age_opts),
    ],
    "platelets": [
        p("150<=x<=450", units=TEN_X_9_PER_LITER, gender=[MALE, FEMALE], **age_opts)
    ],
    "neutrophil": [
        p("2.5<=x<=7.5", units=TEN_X_9_PER_LITER, gender=[MALE, FEMALE], **age_opts)
    ],
    "sodium": [
        p("135<=x<=145", units=MILLIMOLES_PER_LITER, gender=[MALE, FEMALE], **age_opts)
    ],
    "potassium": [
        p("3.6<=x<=5.2", units=MILLIMOLES_PER_LITER, gender=[MALE, FEMALE], **age_opts)
    ],
    "magnesium": [
        p("0.75<=x<=1.2", units=MILLIMOLES_PER_LITER, gender=[MALE, FEMALE], **age_opts)
    ],
    "alt": [p("10<=x<=40", units=IU_LITER, gender=[MALE, FEMALE], **age_opts)],
    "creatinine": [
        p("0.6<=x<=1.3", units=MILLIGRAMS_PER_DECILITER, gender=[MALE, FEMALE], **age_opts),
        p("53<=x<=115", units=MICROMOLES_PER_LITER, gender=[MALE, FEMALE], **age_opts),
    ],
}

grading_data = {
    "haemoglobin": [
        p("7.0<=x<9.0", grade=GRADE3, units=GRAMS_PER_DECILITER, gender=[MALE], **age_opts),
        p("6.5<=x<8.5", grade=GRADE3, units=GRAMS_PER_DECILITER, gender=[FEMALE], **age_opts),
        p("x<7.0", grade=GRADE4, units=GRAMS_PER_DECILITER, gender=[MALE], **age_opts),
        p("x<6.5", grade=GRADE4, units=GRAMS_PER_DECILITER, gender=[FEMALE], **age_opts),
    ],
    "platelets": [
        p(
            "25<=x<=50",
            grade=GRADE3,
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p("x<25", grade=GRADE4, units=TEN_X_9_PER_LITER, gender=[MALE, FEMALE], **age_opts),
    ],
    "neutrophil": [
        p(
            "0.4<=x<=0.59",
            grade=GRADE3,
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p("x<0.4", grade=GRADE4, units=TEN_X_9_PER_LITER, gender=[MALE, FEMALE], **age_opts),
    ],
    "sodium": [
        p(
            "121<=x<=124",
            grade=GRADE3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "154<=x<=159",
            grade=GRADE3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "160<=x",
            grade=GRADE4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "x<=120",
            grade=GRADE4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
    ],
    "potassium": [
        p(
            "2.0<=x<=2.4",
            grade=GRADE3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "6.5<=x<=7.0",
            grade=GRADE3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "x<2.0",
            grade=GRADE4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "7.0<x",
            grade=GRADE4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
    ],
    "magnesium": [
        p(
            "0.3<=x<=0.44",
            grade=GRADE3,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "x<0.3",
            grade=GRADE4,
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
    ],
    "alt": [
        p("200<=x<=400", grade=GRADE3, units=IU_LITER, gender=[MALE, FEMALE], **age_opts),
        p("400<x", grade=GRADE4, units=IU_LITER, gender=[MALE, FEMALE], **age_opts),
    ],
    "creatinine": [
        p(
            "2.47<=x<=4.42",
            grade=GRADE3,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "216<=x<=400",
            grade=GRADE3,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "4.55<x",
            grade=GRADE4,
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
        p(
            "400<x",
            grade=GRADE4,
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **age_opts,
        ),
    ],
}



from edc_lab import LabProfile
from edc_lab_panel.panels import rft_panel

lab_profile = LabProfile(
    name="my_lab_profile",
    requisition_model="egfr_app.subjectrequisition",
    reference_range_collection_name="my_reference_list",
)

lab_profile.add_panel(rft_panel)
